@@ -21,46 +21,25 @@
 #
 
 use strict;
-use Carp;
 
 require DBI;
 DBI->require_version(1.0201);
 
-use RPC::PlClient 0.2000; # XXX change to 0.2017 once it's released
+use RPC::PlClient 0.2000;
 
-{	package DBD::Proxy::RPC::PlClient;
-    	@DBD::Proxy::RPC::PlClient::ISA = qw(RPC::PlClient);
-	sub Call {
-	    my $self = shift;
-	    if ($self->{debug}) {
-		my ($rpcmeth, $obj, $method, @args) = @_;
-		local $^W; # silence undefs
-		Carp::carp("Server $rpcmeth $method(@args)");
-	    }
-	    return $self->SUPER::Call(@_);
-	}
-}
 
 
 package DBD::Proxy;
 
-use vars qw($VERSION $drh %ATTR);
+use vars qw($VERSION $err $errstr $drh);
 
-$VERSION = "0.2004";
 
+$VERSION = "0.2003";
+
+$err = 0;		# holds error code   for DBI::err
+$errstr = "";		# holds error string for DBI::errstr
 $drh = undef;		# holds driver handle once initialised
 
-%ATTR = (	# common to db & st, see also %ATTR in DBD::Proxy::db & ::st
-    'Warn'	=> 'local',
-    'Active'	=> 'local',
-    'Kids'	=> 'local',
-    'CachedKids' => 'local',
-    'PrintError' => 'local',
-    'RaiseError' => 'local',
-    'HandleError' => 'local',
-    'TraceLevel' => 'cached',
-    'CompatMode' => 'local',
-);
 
 sub driver ($$) {
     if (!$drh) {
@@ -71,33 +50,24 @@ sub driver ($$) {
 	$drh = DBI::_new_drh($class, {
 	    'Name' => 'Proxy',
 	    'Version' => $VERSION,
+	    'Err'    => \$DBD::Proxy::err,
+	    'Errstr' => \$DBD::Proxy::errstr,
 	    'Attribution' => 'DBD::Proxy by Jochen Wiedmann',
-	});
-	$drh->STORE(CompatMode => 1); # disable DBI dispatcher attribute cache (for FETCH)
+	    });
     }
     $drh;
 }
 
-sub CLONE {
-    undef $drh;
-}
-
-sub proxy_set_err {
-  my ($h,$errmsg) = @_;
-  my ($err, $state) = ($errmsg =~ s/ \[err=(.*?),state=(.*?)\]//)
-	? ($1, $2) : (1, ' ' x 5);
-  return $h->set_err($err, $errmsg, $state);
-}
 
 package DBD::Proxy::dr; # ====== DRIVER ======
 
 $DBD::Proxy::dr::imp_data_size = 0;
 
 sub connect ($$;$$) {
-    my($drh, $dsn, $user, $auth, $attr)= @_;
+    my($drh, $dsn, $user, $auth)= @_;
     my($dsnOrig) = $dsn;
 
-    my %attr = %$attr;
+    my %attr;
     my ($var, $val);
     while (length($dsn)) {
 	if ($dsn =~ /^dsn=(.*)/) {
@@ -137,56 +107,30 @@ sub connect ($$;$$) {
 	if ($@) { $err .= " Cannot create usercipher object: $@."; }
     }
 
-    return DBD::Proxy::proxy_set_err($drh, $err) if $err; # Returns undef
+    return DBI::set_err($drh, 1, $err) if $err; # Returns undef
 
-    my %client_opts = (
-		       'peeraddr'	=> $attr{'hostname'},
-		       'peerport'	=> $attr{'port'},
-		       'socket_proto'	=> 'tcp',
-		       'application'	=> $attr{dsn},
-		       'user'		=> $user || '',
-		       'password'	=> $auth || '',
-		       'version'	=> $DBD::Proxy::VERSION,
-		       'cipher'	        => $cipherRef,
-		       'debug'		=> $attr{debug}   || 0,
-		       'timeout'	=> $attr{timeout} || undef,
-		       'logfile'	=> $attr{logfile} || undef
-		      );
-    # Options starting with 'proxy_rpc_' are forwarded to the RPC layer after
-    # stripping the prefix.
-    while (my($var,$val) = each %attr) {
-	if ($var =~ s/^proxy_rpc_//) {
-	    $client_opts{$var} = $val;
-	}
-    }
     # Create an RPC::PlClient object.
-    my($client, $msg) = eval { DBD::Proxy::RPC::PlClient->new(%client_opts) };
+    my($client, $msg) = eval { RPC::PlClient->new(
+	'peeraddr'	=> $attr{'hostname'},
+	'peerport'	=> $attr{'port'},
+	'socket_proto'	=> 'tcp',
+	'application'	=> $attr{dsn},
+	'user'		=> $user || '',
+	'password'	=> $auth || '',
+	'version'	=> $DBD::Proxy::VERSION,
+	'cipher'	=> $cipherRef,
+	'debug'		=> $attr{debug}   || 0,
+	'timeout'	=> $attr{timeout} || undef,
+	'logfile'	=> $attr{logfile} || undef
+    ) };
 
-    return DBD::Proxy::proxy_set_err($drh, "Cannot log in to DBI::ProxyServer: $@")
+    return DBI::set_err($drh, 1, "Cannot log in to DBI::ProxyServer: $@")
 	if $@; # Returns undef
-    return DBD::Proxy::proxy_set_err($drh, "Constructor didn't return a handle: $msg")
+    return DBI::set_err($drh, 1, "Constructor didn't return a handle: $msg")
 	unless ($msg =~ /^((?:\w+|\:\:)+)=(\w+)/); # Returns undef
 
     $msg = RPC::PlClient::Object->new($1, $client, $msg);
 
-    my $max_proto_ver;
-    my ($server_ver_str) = eval { $client->Call('Version') };
-    if ( $@ ) {
-      # Server denies call, assume legacy protocol.
-      $max_proto_ver = 1;
-    } else {
-      # Parse proxy server version.
-      my ($server_ver_num) = $server_ver_str =~ /^DBI::ProxyServer\s+([\d\.]+)/;
-      $max_proto_ver = $server_ver_num >= 0.3 ? 2 : 1;
-    }
-    my $req_proto_ver;
-    if ( exists $attr{proxy_lazy_prepare} ) {
-      $req_proto_ver = ($attr{proxy_lazy_prepare} == 0) ? 2 : 1;
-      return DBD::Proxy::proxy_set_err($drh, 
-                 "DBI::ProxyServer does not support synchronous statement preparation.")
-	if $max_proto_ver < $req_proto_ver;
-    }
-
     # Switch to user specific encryption mode, if desired
     if ($userCipherRef) {
 	$client->{'cipher'} = $userCipherRef;
@@ -197,8 +141,7 @@ sub connect ($$;$$) {
 	    'Name' => $dsnOrig,
 	    'proxy_dbh' => $msg,
 	    'proxy_client' => $client,
-	    'RowCacheSize' => $attr{'RowCacheSize'} || 20,
-	    'proxy_proto_ver' => $req_proto_ver || 1
+	    'RowCacheSize' => $attr{'RowCacheSize'} || 20
    });
 
     foreach $var (keys %attr) {
@@ -206,12 +149,13 @@ sub connect ($$;$$) {
 	    $this->{$var} = $attr{$var};
 	}
     }
-    $this->SUPER::STORE('Active' => 1);
 
     $this;
 }
 
 
+sub disconnect_all { }
+
 sub DESTROY { undef }
 
 
@@ -219,39 +163,16 @@ package DBD::Proxy::db; # ====== DATABASE ======
 
 $DBD::Proxy::db::imp_data_size = 0;
 
-# XXX probably many more methods need to be added here
-# in order to trigger our AUTOLOAD to redirect them to the server.
-# (Unless the sub is declared it's bypassed by perl method lookup.)
-# See notes in ToDo about method metadata
-# The question is whether to add all the methods in %DBI::DBI_methods
-# to the corresponding classes (::db, ::st etc)
-# Also need to consider methods that, if proxied, would change the server state
-# in a way that might not be visible on the client, ie begin_work -> AutoCommit.
-
-sub commit;
-sub connected;
-sub rollback;
-sub ping;
-
-
 use vars qw(%ATTR $AUTOLOAD);
 
-# inherited: STORE / FETCH against this class.
-# local:     STORE / FETCH against parent class.
-# cached:    STORE to remote and local objects, FETCH from local.
-# remote:    STORE / FETCH against remote object only (default).
-#
-# Note: Attribute names starting with 'proxy_' always treated as 'inherited'.
-#
-%ATTR = (	# see also %ATTR in DBD::Proxy::st
-    %DBD::Proxy::ATTR,
-    RowCacheSize => 'inherited',
-    #AutoCommit => 'cached',
-    'FetchHashKeyName' => 'cached',
-    Statement => 'local',
-    Driver => 'local',
-    dbi_connect_closure => 'local',
-    Username => 'local',
+%ATTR = (
+    'Warn' => 'local',
+    'Active' => 'local',
+    'Kids' => 'local',
+    'CachedKids' => 'local',
+    'PrintError' => 'local',
+    'RaiseError' => 'local',
+    'RowCacheSize' => 'inherited'
 );
 
 sub AUTOLOAD {
@@ -259,66 +180,47 @@ sub AUTOLOAD {
     $method =~ s/(.*::(.*)):://;
     my $class = $1;
     my $type = $2;
-    #warn "AUTOLOAD of $method (class=$class, type=$type)";
-    my %expand = (
-        'method' => $method,
-        'class' => $class,
-        'type' => $type,
-        'call' => "$method(\@_)",
-        # XXX was trying to be smart but was tripping up over the DBI's own
-        # smartness. Disabled, but left here in case there are issues.
-    #   'call' => (UNIVERSAL::can("DBI::_::$type", $method)) ? "$method(\@_)" : "func(\@_, '$method')",
-    );
-
-    my $method_code = q{
-        package ~class~;
-        sub ~method~ {
+    my %expand =
+	( 'method' => $method,
+	  'class' => $class,
+	  'type' => $type,
+	  'h' => "DBI::_::$type"
+	);
+    my $method_code = UNIVERSAL::can($expand{'h'}, $method) ?
+	q/package ~class~;
+          sub ~method~ {
             my $h = shift;
-            local $@;
-            my @result = wantarray
-                ? eval {        $h->{'proxy_~type~h'}->~call~ }
-                : eval { scalar $h->{'proxy_~type~h'}->~call~ };
-            return DBD::Proxy::proxy_set_err($h, $@) if $@;
-            return wantarray ? @result : $result[0];
-        }
-    };
+	    my @result = eval { $h->{'proxy_~type~h'}->~method~(@_) };
+            return DBI::set_err($h, 1, $@) if $@;
+            wantarray ? @result : $result[0];
+          }
+	 / :
+        q/package ~class~;
+	  sub ~method~ {
+	    my $h = shift;
+	    my @result = eval { $h->{'proxy_~type~h'}->func(@_, '~method~') };
+	    return DBI::set_err($h, 1, $@) if $@;
+	    wantarray ? @result : $result[0];
+          }
+         /;
     $method_code =~ s/\~(\w+)\~/$expand{$1}/eg;
-    local $SIG{__DIE__} = 'DEFAULT';
-    my $err = do { local $@; eval $method_code.2; $@ };
-    die $err if $err;
+    eval $method_code;
+    die $@ if $@;
     goto &$AUTOLOAD;
 }
 
 sub DESTROY {
-    my $dbh = shift;
-    local $@ if $@;	# protect $@
-    $dbh->disconnect if $dbh->SUPER::FETCH('Active');
+    # Just to avoid that DESTROY is autoloaded ...
 }
 
 sub disconnect ($) {
-    my ($dbh) = @_;
-
-    # Sadly the Proxy too-often disagrees with the backend database
-    # on the subject of 'Active'.  In the short term, I'd like the
-    # Proxy to ease up and let me decide when it's proper to go over
-    # the wire.  This ultimately applies to finish() as well.
-    #return unless $dbh->SUPER::FETCH('Active');
-
-    # Drop database connection at remote end
-    my $rdbh = $dbh->{'proxy_dbh'};
-    if ( $rdbh ) {
-        local $SIG{__DIE__} = 'DEFAULT';
-        local $@;
-	eval { $rdbh->disconnect() } ;
-        DBD::Proxy::proxy_set_err($dbh, $@) if $@;
-    }
-    
-    # Close TCP connect to remote
-    # XXX possibly best left till DESTROY? Add a config attribute to choose?
-    #$dbh->{proxy_client}->Disconnect(); # Disconnect method requires newer PlRPC module
-    $dbh->{proxy_client}->{socket} = undef; # hack
-
-    $dbh->SUPER::STORE('Active' => 0);
+    my($dbh) = @_;
+    # XXX this should call $rdbh->disconnect to get the right
+    # disconnect behaviour. It should not undef these values.
+    # A proxy_no_disconnect option could be added (like for finish)
+    # to let people trade safety for speed if they need to.
+    undef $dbh->{'proxy_dbh'};    # Bug in Perl 5.004; would prefer delete
+    undef $dbh->{'proxy_client'};
     1;
 }
 
@@ -327,25 +229,14 @@ sub STORE ($$$) {
     my($dbh, $attr, $val) = @_;
     my $type = $ATTR{$attr} || 'remote';
 
-    if ($attr eq 'TraceLevel') {
-	warn("TraceLevel $val");
-	my $pc = $dbh->{proxy_client} || die;
-	$pc->{logfile} ||= 1; # XXX hack
-	$pc->{debug} = ($val && $val >= 4);
-	$pc->Debug("$pc debug enabled") if $pc->{debug};
-    }
-
     if ($attr =~ /^proxy_/  ||  $type eq 'inherited') {
 	$dbh->{$attr} = $val;
 	return 1;
     }
 
-    if ($type eq 'remote' ||  $type eq 'cached') {
-        local $SIG{__DIE__} = 'DEFAULT';
-	local $@;
+    if ($type eq 'remote') {
 	my $result = eval { $dbh->{'proxy_dbh'}->STORE($attr => $val) };
-	return DBD::Proxy::proxy_set_err($dbh, $@) if $@; # returns undef
-	$dbh->SUPER::STORE($attr => $val) if $type eq 'cached';
+	return DBI::set_err($dbh, 1, $@) if $@; # returns undef
 	return $result;
     }
     return $dbh->SUPER::STORE($attr => $val);
@@ -353,52 +244,31 @@ sub STORE ($$$) {
 
 sub FETCH ($$) {
     my($dbh, $attr) = @_;
-    # we only get here for cached attribute values if the handle is in CompatMode
-    # otherwise the DBI dispatcher handles the FETCH itself from the attribute cache.
     my $type = $ATTR{$attr} || 'remote';
 
-    if ($attr =~ /^proxy_/  ||  $type eq 'inherited'  || $type eq 'cached') {
+    if ($attr =~ /^proxy_/  ||  $type eq 'inherited') {
 	return $dbh->{$attr};
     }
 
     return $dbh->SUPER::FETCH($attr) unless $type eq 'remote';
 
-    local $SIG{__DIE__} = 'DEFAULT';
-    local $@;
     my $result = eval { $dbh->{'proxy_dbh'}->FETCH($attr) };
-    return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
+    return DBI::set_err($dbh, 1, $@) if $@;
     return $result;
 }
 
 sub prepare ($$;$) {
     my($dbh, $stmt, $attr) = @_;
+
+    # We *could* send the statement over the net immediately, but the
+    # DBI specs allows us to defer that until the first 'execute'.
+    # XXX should make this configurable
     my $sth = DBI::_new_sth($dbh, {
-				   'Statement' => $stmt,
-				   'proxy_attr' => $attr,
-				   'proxy_cache_only' => 0,
-				   'proxy_params' => [],
-				  }
-			   );
-    my $proto_ver = $dbh->{'proxy_proto_ver'};
-    if ( $proto_ver > 1 ) {
-      $sth->{'proxy_attr_cache'} = {cache_filled => 0};
-      my $rdbh = $dbh->{'proxy_dbh'};
-      local $SIG{__DIE__} = 'DEFAULT';
-      local $@;
-      my $rsth = eval { $rdbh->prepare($sth->{'Statement'}, $sth->{'proxy_attr'}, undef, $proto_ver) };
-      return DBD::Proxy::proxy_set_err($sth, $@) if $@;
-      return DBD::Proxy::proxy_set_err($sth, "Constructor didn't return a handle: $rsth")
-	unless ($rsth =~ /^((?:\w+|\:\:)+)=(\w+)/);
-    
-      my $client = $dbh->{'proxy_client'};
-      $rsth = RPC::PlClient::Object->new($1, $client, $rsth);
-      
-      $sth->{'proxy_sth'} = $rsth;
-      # If statement is a positioned update we do not want any readahead.
-      $sth->{'RowCacheSize'} = 1 if $stmt =~ /\bfor\s+update\b/i;
-    # Since resources are used by prepared remote handle, mark us active.
-    $sth->SUPER::STORE(Active => 1);
-    }
+	    'Statement' => $stmt,
+	    'proxy_attr' => $attr,
+	    'proxy_params' => [],
+    	'proxy_cache_only' => 0,
+    });
     $sth;
 }
 
@@ -422,22 +292,17 @@ sub quote {
     #    $dbh->{'proxy_quote'} = 'backslash_escaped';
     # for example.
     # Jochen
-    local $SIG{__DIE__} = 'DEFAULT';
-    local $@;
+
     my $result = eval { $dbh->{'proxy_dbh'}->quote(@_) };
-    return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
+    return DBI::set_err($dbh, 1, $@) if $@;
     return $result;
 }
 
 sub table_info {
     my $dbh = shift;
     my $rdbh = $dbh->{'proxy_dbh'};
-    #warn "table_info(@_)";
-    local $SIG{__DIE__} = 'DEFAULT';
-    local $@;
     my($numFields, $names, $types, @rows) = eval { $rdbh->table_info(@_) };
-    return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
-    my ($sth, $inner) = DBI::_new_sth($dbh, {
+    my $sth = DBI::_new_sth($dbh, {
         'Statement' => "SHOW TABLES",
 	'proxy_params' => [],
 	'proxy_data' => \@rows,
@@ -445,32 +310,18 @@ sub table_info {
 		'NUM_OF_PARAMS' => 0, 
 		'NUM_OF_FIELDS' => $numFields, 
 		'NAME' => $names, 
-		'TYPE' => $types,
-		'cache_filled' => 1
+		'TYPE' => $types
 		},
     	'proxy_cache_only' => 1,
     });
     $sth->SUPER::STORE('NUM_OF_FIELDS' => $numFields);
-    $inner->{NAME} = $names;
-    $inner->{TYPE} = $types;
-    $sth->SUPER::STORE('Active' => 1); # already execute()'d
-    $sth->{'proxy_rows'} = @rows;
     return $sth;
 }
 
-sub tables {
-    my $dbh = shift;
-    #warn "tables(@_)";
-    return $dbh->SUPER::tables(@_);
-}
-
-
 sub type_info_all {
     my $dbh = shift;
-    local $SIG{__DIE__} = 'DEFAULT';
-    local $@;
     my $result = eval { $dbh->{'proxy_dbh'}->type_info_all(@_) };
-    return DBD::Proxy::proxy_set_err($dbh, $@) if $@;
+    return DBI::set_err($dbh, 1, $@) if $@;
     return $result;
 }
 
@@ -481,18 +332,13 @@ $DBD::Proxy::st::imp_data_size = 0;
 
 use vars qw(%ATTR);
 
-# inherited:  STORE to current object. FETCH from current if exists, else call up
-#              to the (proxy) database object.
-# local:      STORE / FETCH against parent class.
-# cache_only: STORE noop (read-only).  FETCH from private_* if exists, else call
-#              remote and cache the result.
-# remote:     STORE / FETCH against remote object only (default).
-#
-# Note: Attribute names starting with 'proxy_' always treated as 'inherited'.
-#
-%ATTR = (	# see also %ATTR in DBD::Proxy::db
-    %DBD::Proxy::ATTR,
-    'Database' => 'local',
+%ATTR = (
+    'Warn' => 'local',
+    'Active' => 'local',
+    'Kids' => 'local',
+    'CachedKids' => 'local',
+    'PrintError' => 'local',
+    'RaiseError' => 'local',
     'RowsInCache' => 'local',
     'RowCacheSize' => 'inherited',
     'NULLABLE' => 'cache_only',
@@ -512,74 +358,61 @@ sub execute ($@) {
 
     # new execute, so delete any cached rows from previous execute
     undef $sth->{'proxy_data'};
-    undef $sth->{'proxy_rows'};
 
+    my $dbh = $sth->{'Database'};
+    my $client = $dbh->{'proxy_client'};
     my $rsth = $sth->{proxy_sth};
-    my $dbh = $sth->FETCH('Database');
-    my $proto_ver = $dbh->{proxy_proto_ver};
 
-    my ($numRows, @outData);
+    my ($numFields, $numParams, $numRows, $names, $types, @outParams);
 
-    local $SIG{__DIE__} = 'DEFAULT';
-    local $@;
-    if ( $proto_ver > 1 ) {
-      ($numRows, @outData) = eval { $rsth->execute($params, $proto_ver) };
-      return DBD::Proxy::proxy_set_err($sth, $@) if $@;
-      
-      # Attributes passed back only on the first execute() of a statement.
-      unless ($sth->{proxy_attr_cache}->{cache_filled}) {
-	my ($numFields, $numParams, $names, $types) = splice(@outData, 0, 4); 
-	$sth->{'proxy_attr_cache'} = {
-				      'NUM_OF_FIELDS' => $numFields,
-				      'NUM_OF_PARAMS' => $numParams,
-				      'NAME'          => $names,
-				      'cache_filled'  => 1
-				     };
-	$sth->SUPER::STORE('NUM_OF_FIELDS' => $numFields);
-	$sth->SUPER::STORE('NUM_OF_PARAMS' => $numParams);
-      }
+    if ($sth->{'proxy_data'}) {
+	my $attrCache = $sth->{'proxy_attr_cache'};
+	$numFields = $attrCache->{'NUM_OF_FIELDS'};
+	$numParams = $attrCache->{'NUM_OF_PARAMS'};
+	$names = $attrCache->{'NAME'};
+	$types = $attrCache->{'TYPE'};
+	$numRows = scalar @{$sth->{'proxy_data'}};
+    } else {
 
-    }
-    else {
-      if ($rsth) {
-	($numRows, @outData) = eval { $rsth->execute($params, $proto_ver) };
-	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
-
-      }
-      else {
+    if (!$rsth) {
 	my $rdbh = $dbh->{'proxy_dbh'};
-	
-	# Legacy prepare is actually prepare + first execute on the server.
-        ($rsth, @outData) =
-	  eval { $rdbh->prepare($sth->{'Statement'},
-				$sth->{'proxy_attr'}, $params, $proto_ver) };
-	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
-	return DBD::Proxy::proxy_set_err($sth, "Constructor didn't return a handle: $rsth")
-	  unless ($rsth =~ /^((?:\w+|\:\:)+)=(\w+)/);
-	
-	my $client = $dbh->{'proxy_client'};
+
+	($rsth, $numFields, $numParams, $names, $types, $numRows, @outParams) =
+	    eval { $rdbh->prepare($sth->{'Statement'},
+				  $sth->{'proxy_attr'}, $params) };
+	return DBI::set_err($sth, 1, $@) if $@;
+	return DBI::set_err($sth, 1,
+			    "Constructor didn't return a handle: $rsth")
+	    unless ($rsth =~ /^((?:\w+|\:\:)+)=(\w+)/);
+
 	$rsth = RPC::PlClient::Object->new($1, $client, $rsth);
 
-	my ($numFields, $numParams, $names, $types) = splice(@outData, 0, 4);
 	$sth->{'proxy_sth'} = $rsth;
-        $sth->{'proxy_attr_cache'} = {
+ 	$sth->SUPER::STORE('NUM_OF_FIELDS' => $numFields);
+ 	$sth->SUPER::STORE('NUM_OF_PARAMS' => $numParams);
+    } else {
+	my $attrCache = $sth->{'proxy_attr_cache'};
+	$numFields = $attrCache->{'NUM_OF_FIELDS'};
+	$numParams = $attrCache->{'NUM_OF_PARAMS'};
+	$names = $attrCache->{'NAME'};
+	$types = $attrCache->{'TYPE'};
+	($numRows, @outParams) = eval { $rsth->execute($params) };
+	return DBI::set_err($sth, 1, $@) if $@;
+    }
+    }
+    $sth->{'proxy_rows'} = $numRows;
+    $sth->{'proxy_attr_cache'} = {
 	    'NUM_OF_FIELDS' => $numFields,
 	    'NUM_OF_PARAMS' => $numParams,
 	    'NAME'          => $names
-        };
-	$sth->SUPER::STORE('NUM_OF_FIELDS' => $numFields);
-	$sth->SUPER::STORE('NUM_OF_PARAMS' => $numParams);
-	$numRows = shift @outData;
-      }
-    }
-    # Always condition active flag.
-    $sth->SUPER::STORE('Active' => 1) if $sth->FETCH('NUM_OF_FIELDS'); # is SELECT
-    $sth->{'proxy_rows'} = $numRows;
-    # Any remaining items are output params.
-    if (@outData) {
+    };
+
+    $sth->SUPER::STORE('Active' => 1) if $numFields; # is SELECT
+
+    if (@outParams) {
 	foreach my $p (@$params) {
-	    if (ref($p->[0])) {
-		my $ref = shift @outData;
+	    if (ref($p)  &&  @$p > 2) {
+		my $ref = shift @outParams;
 		${$p->[0]} = $$ref;
 	    }
 	}
@@ -593,9 +426,7 @@ sub fetch ($) {
 
     my $data = $sth->{'proxy_data'};
 
-    $sth->{'proxy_rows'} = 0 unless defined $sth->{'proxy_rows'};
-
-    if(!$data || !@$data) {
+    if(!$data  ||  !@$data) {
 	return undef unless $sth->SUPER::FETCH('Active');
 
 	my $rsth = $sth->{'proxy_sth'};
@@ -603,10 +434,8 @@ sub fetch ($) {
 	    die "Attempt to fetch row without execute";
 	}
 	my $num_rows = $sth->FETCH('RowCacheSize') || 20;
-	local $SIG{__DIE__} = 'DEFAULT';
-	local $@;
 	my @rows = eval { $rsth->fetch($num_rows) };
-	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
+	return DBI::set_err($sth, 1, $@) if $@;
 	unless (@rows == $num_rows) {
 	    undef $sth->{'proxy_data'};
 	    # server side has already called finish
@@ -617,15 +446,14 @@ sub fetch ($) {
     }
     my $row = shift @$data;
 
-    $sth->SUPER::STORE(Active => 0) if ( $sth->{proxy_cache_only} and !@$data );
-    $sth->{'proxy_rows'}++;
+	$sth->SUPER::STORE(Active => 0) if ( $sth->{proxy_cache_only} and !@$data );
     return $sth->_set_fbav($row);
 }
 *fetchrow_arrayref = \&fetch;
 
 sub rows ($) {
-    my $rows = shift->{'proxy_rows'};
-    return (defined $rows) ? $rows : -1;
+    my($sth) = @_;
+    $sth->{'proxy_rows'};
 }
 
 sub finish ($) {
@@ -636,12 +464,10 @@ sub finish ($) {
     return 0 unless $rsth; # Something's out of sync
     my $no_finish = exists($sth->{'proxy_no_finish'})
  	? $sth->{'proxy_no_finish'}
-	: $sth->FETCH('Database')->{'proxy_no_finish'};
+	: $sth->{Database}->{'proxy_no_finish'};
     unless ($no_finish) {
-        local $SIG{__DIE__} = 'DEFAULT';
-	local $@;
 	my $result = eval { $rsth->finish() };
-	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
+	return DBI::set_err($sth, 1, $@) if $@;
 	return $result;
     }
     1;
@@ -660,13 +486,11 @@ sub STORE ($$$) {
 	return 0;
     }
 
-    if ($type eq 'remote' || $type eq 'cached') {
+    if ($type eq 'remote') {
 	my $rsth = $sth->{'proxy_sth'}  or  return undef;
-        local $SIG{__DIE__} = 'DEFAULT';
-	local $@;
 	my $result = eval { $rsth->STORE($attr => $val) };
-	return DBD::Proxy::proxy_set_err($sth, $@) if ($@);
-	return $result if $type eq 'remote'; # else fall through to cache locally
+	return DBI::set_err($sth, 1, $@) if ($@);
+	return $result;
     }
     return $sth->SUPER::STORE($attr => $val);
 }
@@ -683,27 +507,23 @@ sub FETCH ($$) {
 	if (exists($sth->{$attr})) {
 	    return $sth->{$attr};
 	}
-	return $sth->FETCH('Database')->{$attr};
+	return $sth->{'Database'}->{$attr};
     }
 
     if ($type eq 'cache_only'  &&
-	    exists($sth->{'proxy_attr_cache'}->{$attr})) {
+	exists($sth->{'proxy_attr_cache'}->{$attr})) {
 	return $sth->{'proxy_attr_cache'}->{$attr};
     }
 
     if ($type ne 'local') {
 	my $rsth = $sth->{'proxy_sth'}  or  return undef;
-        local $SIG{__DIE__} = 'DEFAULT';
-	local $@;
 	my $result = eval { $rsth->FETCH($attr) };
-	return DBD::Proxy::proxy_set_err($sth, $@) if $@;
+	return DBI::set_err($sth, 1, $@) if $@;
 	return $result;
-    }
-    elsif ($attr eq 'RowsInCache') {
+    } elsif ($attr eq 'RowsInCache') {
 	my $data = $sth->{'proxy_data'};
 	$data ? @$data : 0;
-    }
-    else {
+    } else {
 	$sth->SUPER::FETCH($attr);
     }
 }
@@ -715,8 +535,7 @@ sub bind_param ($$$@) {
 *bind_param_inout = \&bind_param;
 
 sub DESTROY {
-    my $sth = shift;
-    $sth->finish if $sth->SUPER::FETCH('Active');
+    # Just to avoid autoloading DESTROY ...
 }
 
 
@@ -766,7 +585,7 @@ I<Note that the connect_cached method is new and still experimental.>
 Before connecting to a remote database, you must ensure, that a Proxy
 server is running on the remote machine. There's no default port, so
 you have to ask your system administrator for the port number. See
-L<DBI::ProxyServer> for details.
+L<DBI::ProxyServer(3)> for details.
 
 Say, your Proxy server is running on machine "alpha", port 3334, and
 you'd like to connect to an ODBC database called "mydb" as user "joe"
@@ -783,7 +602,7 @@ You see, this is mainly the same. The DBD::Proxy module will create a
 connection to the Proxy server on "alpha" which in turn will connect
 to the ODBC database.
 
-Refer to the L<DBI> documentation on the C<connect> method for a way
+Refer to the L<DBI(3)> documentation on the C<connect> method for a way
 to automatically use DBD::Proxy without having to change your code.
 
 DBD::Proxy's DSN string has the format
@@ -833,7 +652,7 @@ by executing
     $cipherRef = $class->new(pack("H*", $key));
 
 and pass this object to the RPC::PlClient module when creating a
-client. See L<RPC::PlClient>. Example:
+client. See L<RPC::PlClient(3)>. Example:
 
     cipher=IDEA;key=97cd2375efa329aceef2098babdc9721
 
@@ -845,7 +664,7 @@ less secure than the usercipher/userkey secret and readable by anyone.
 The usercipher/userkey secret is B<your> private secret.
 
 Of course encryption requires an appropriately configured server. See
-<DBD::ProxyServer/CONFIGURATION FILE>.
+<DBD::ProxyServer(3)/CONFIGURATION FILE>.
 
 =item debug
 
@@ -892,7 +711,7 @@ and fast CGI applications.
 This attribute can be used to reduce network traffic: By default calls
 to $dbh->quote() are passed to the remote driver.  Of course this slows
 down things quite a lot, but is the safest default behaviour.
-
+  
 However, if you set the I<proxy_quote> attribute to the value 'C<local>'
 either in the database handle or in the statement handle, and the call
 to quote has only one parameter, then the local default DBI quote
@@ -902,26 +721,6 @@ method will be used (which will be faster but may be wrong).
 
 =head1 KNOWN ISSUES
 
-=head2 Unproxied method calls
-
-If a method isn't being proxied, try declaring a stub sub in the appropriate
-package (DBD::Proxy::db for a dbh method, and DBD::Proxy::st for an sth method).
-For example:
-
-    sub DBD::Proxy::db::selectall_arrayref;
-
-That will enable selectall_arrayref to be proxied.
-
-Currently many methods aren't explicitly proxied and so you get the DBI's
-default methods executed on the client.
-
-Some of those methods, like selectall_arrayref, may then call other methods
-that are proxied (selectall_arrayref calls fetchall_arrayref which calls fetch
-which is proxied). So things may appear to work but operate more slowly than
-the could.
-
-This may all change in a later version.
-
 =head2 Complex handle attributes
 
 Sometimes handles are having complex attributes like hash refs or
@@ -942,13 +741,13 @@ executing the above example in two steps:
 
 =over
 
-=item 1
-
+=item 1.)
+ 
 The first step is fetching the value of the key "csv_tables" in the
 handle $dbh. The value returned is complex, a hash ref.
-
-=item 2
-
+ 
+=item 2.)
+ 
 The second step is storing some value (the right hand side of the
 assignment) as the key "passwd" in the hash ref from step 1.
 
@@ -992,6 +791,6 @@ is granted to Tim Bunce for distributing this as a part of the DBI.
 
 =head1 SEE ALSO
 
-L<DBI>, L<RPC::PlClient>, L<Storable>
+L<DBI(3)>, L<RPC::PlClient(3)>, L<Storable(3)>
 
 =cut
