@@ -30,6 +30,13 @@ require Config;
 package DBI::ProxyServer;
 
 
+my $haveFileSpec = eval { require File::Spec };
+my $tmpDir = $haveFileSpec ? File::Spec->tmpdir() :
+    ($ENV{'TMP'} || $ENV{'TEMP'} || '/tmp');
+my $defaultPidFile = $haveFileSpec ?
+    File::Spec->catdir($tmpDir, "dbiproxy.pid") : "/tmp/dbiproxy.pid";
+
+
 ############################################################################
 #
 #   Constants
@@ -95,7 +102,7 @@ my %DEFAULT_SERVER_OPTIONS;
     } else {
 	$o->{'mode'} = 'single';
     }
-    $o->{'pidfile'}    = '/tmp/dbiproxy.pid';
+    $o->{'pidfile'}    = $defaultPidFile;
     $o->{'user'}       = undef;
 };
 
@@ -171,6 +178,7 @@ sub AcceptVersion {
 
 sub AcceptUser {
     my $self = shift; my $user = shift; my $password = shift;
+    return 0 if (!$self->SUPER::AcceptUser($user, $password));
     my $dsn = $self->{'application'};
     $self->Debug("Connecting to $dsn as $user");
     local $ENV{DBI_AUTOPROXY} = ''; # :-)
